#

from Products.Archetypes.ReferenceEngine import Reference
from Products.PloneTestCase import five
class CustomReference( Reference ): pass
        
class IterateLayer:

    def setUp(cls):
        '''Sets up the CA by loading etc/site.zcml.'''
        five.safe_load_site()
    setUp = classmethod(setUp)

    def tearDown(cls):
        '''Cleans up the CA.'''
        five.cleanUp()
    tearDown = classmethod(tearDown)



##################################################################
#
# (C) Copyright 2006-2007 ObjectRealms, LLC
# All Rights Reserved
#
# This file is part of iterate.
#
# iterate is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# iterate is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with CMFDeployment; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
##################################################################
"""
$Id: workflow.py 1824 2007-02-08 17:59:41Z hazmat $

Applies new checkout specific workflows to content that is checked out.

"""

from Acquisition import aq_base
from Products.CMFCore.utils import getToolByName
from Products.CMFPlacefulWorkflow.WorkflowPolicyConfig import WorkflowPolicyConfig
from Products.CMFPlacefulWorkflow.PlacefulWorkflowTool import WorkflowPolicyConfig_id
from plone.app.iterate.util import get_storage

USE_WORKFLOW = "checkout_workflow_policy"

policy_storage = "previous_wf_policy"

def handleCheckout( event ):
    # defer to setting
    properties = getToolByName(aq_base(event.object), 'portal_properties')
    enabled = properties.site_properties.getProperty('enable_checkout_workflow')
    if not enabled:
        return

    policy_id = properties.site_properties.getProperty('checkout_workflow_policy')
    
    existing_policy = getattr( aq_base( event.working_copy ), WorkflowPolicyConfig_id, None )
    storage = get_storage( event.relation )
    
    # set config for policy in and below 
    policy = WorkflowPolicyConfig( policy_id, policy_id)
    policy.coci_created = True
    
    if existing_policy is not None:
        storage[ policy_storage ] = policy

    # we setattr because we want the effect on non containerish objects
    setattr( event.working_copy, WorkflowPolicyConfig_id, policy )

def handleCheckin( event ):
    policy = getattr( aq_base(event.object), WorkflowPolicyConfig_id, None )
    storage = get_storage( event.relation )
    previous_policy = storage.get( policy_storage )
    if previous_policy is None:
        if policy and not policy.coci_created:  # only reset workflows we know.. could use are own storage
            return
        elif policy is None:
            return
        else:
            delattr( event.object, WorkflowPolicyConfig_id )
    else:
        setattr( event.object, WorkflowPolicyConfig_id, previous_policy)
    


    



"""
$Id: base.py 1808 2007-02-06 11:39:11Z hazmat $
"""

from zope.interface import implements

from zope.viewlet.interfaces import IViewlet

from DateTime import DateTime
from AccessControl import getSecurityManager

from Products.Five.browser import BrowserView
from Products.Five.browser.pagetemplatefile import ViewPageTemplateFile
from Products.CMFCore.permissions import ModifyPortalContent
from Products.CMFCore.utils import getToolByName

from plone.app.iterate.util import get_storage
from plone.app.iterate.interfaces import keys

from plone.app.iterate.relation import WorkingCopyRelation

from plone.memoize.instance import memoize

class BaseInfoViewlet( BrowserView ):
    
    implements( IViewlet )
    
    def __init__( self, context, request, view, manager ):
        super( BaseInfoViewlet, self ).__init__( context, request )
        self.__parent__ = view
        self.view = view
        self.manager = manager
        
    def update( self ):
        pass
    
    def render( self ):
        raise NotImplementedError
        
    @memoize
    def created( self ):
        time = self.properties.get( keys.checkout_time, DateTime() )
        util = getToolByName(self.context, 'translation_service')
        return util.ulocalized_time(time, None, self.context, domain='plonelocales')

    @memoize
    def creator( self ):
        user_id = self.properties.get( keys.checkout_user )
        membership = getToolByName(self.context, 'portal_membership')
        if not user_id:
            return membership.getAuthenticatedMember()
        return membership.getMemberById( user_id )
        
    @memoize
    def creator_url( self ):
        creator = self.creator()
        portal_url = getToolByName(self.context, 'portal_url')
        return "%s/author/%s" % ( portal_url(), creator.getId() )
        
    @memoize
    def creator_name( self ):
        creator = self.creator()
        return creator.getProperty('fullname') or creator.getId()

    @property
    @memoize
    def properties( self ):
        wc_ref = self._getReference()
        if wc_ref is not None:
            return get_storage( wc_ref )
        else:
            return {}

    def _getReference( self ):
        raise NotImplemented
        
class BaselineInfoViewlet( BaseInfoViewlet ):
    
    template = ViewPageTemplateFile('info_baseline.pt')

    def render(self):
        if self.working_copy() is not None and \
            getSecurityManager().checkPermission(ModifyPortalContent, self.context):
            return self.template()
        else:
            return ""

    @memoize
    def working_copy( self ):
        refs = self.context.getBRefs( WorkingCopyRelation.relationship )
        if len( refs ) > 0:
            return refs[0]
        else:
            return None

    def _getReference( self ):
        refs = self.context.getBackReferenceImpl( WorkingCopyRelation.relationship )
        if len( refs ) > 0:
            return refs[0]
        else:
            return None
        
class CheckoutInfoViewlet( BaseInfoViewlet ):
    
    template = ViewPageTemplateFile('info_checkout.pt')
    
    def render(self):
        if self.baseline() is not None and \
            getSecurityManager().checkPermission(ModifyPortalContent, self.context):
            return self.template()
        else:
            return ""
    
    @memoize
    def baseline( self ):
        refs = self.context.getReferences( WorkingCopyRelation.relationship )
        if len( refs ) > 0:
            return refs[0]
        else:
            return None
    
    def _getReference( self ):
        refs = self.context.getReferenceImpl( WorkingCopyRelation.relationship )
        if len( refs ) > 0:
            return refs[0]
        else:
            return None
        



##################################################################
#
# (C) Copyright 2006 ObjectRealms, LLC
# All Rights Reserved
#
# This file is part of iterate.
#
# iterate is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# iterate is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with CMFDeployment; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
##################################################################
"""
$Id: event.py 1811 2007-02-06 18:40:02Z hazmat $
"""
from zope.interface import implements
from zope.event import notify
from zope.component.interfaces import ObjectEvent

import interfaces
import relation

class CheckoutEvent( ObjectEvent ):

    implements( interfaces.ICheckoutEvent )

    def __init__(self, baseline, wc, relation):
        ObjectEvent.__init__(self, baseline )
        self.working_copy = wc
        self.relation = relation
        
class CheckinEvent( ObjectEvent ):

    implements( interfaces.ICheckinEvent )

    def __init__(self, wc, baseline, relation, message):
        ObjectEvent.__init__( self, wc )
        self.baseline = baseline
        self.relation = relation
        self.message = message

class AfterCheckinEvent( ObjectEvent ):

    implements( interfaces.IAfterCheckinEvent )
    
    def __init__( self, new_baseline, checkin_message ):
        super( AfterCheckinEvent, self).__init__( new_baseline )
        self.message = checkin_message

class CancelCheckoutEvent( ObjectEvent ):

    implements( interfaces.ICancelCheckoutEvent )

    def __init__( self, wc, baseline):
        ObjectEvent.__init__(self, wc )
        self.baseline = baseline

class WorkingCopyDeletedEvent( ObjectEvent ):

    implements( interfaces.IWorkingCopyDeletedEvent )

    def __init__( self, wc, baseline, relation ):
        ObjectEvent.__init__( self, wc )
        self.baseline = baseline
        self.relation = relation

class BeforeCheckoutEvent( ObjectEvent ):

    implements( interfaces.IBeforeCheckoutEvent )
    

def handleDeletion( reference, event ):
    # a filtering/enriching event rebroadcaster for working copy deletions
    workingCopy = reference.getSourceObject()
    baseline = reference.getTargetObject()
    notify( WorkingCopyDeletedEvent( workingCopy, baseline, reference ) )
    



##################################################################
#
# (C) Copyright 2006 ObjectRealms, LLC
# All Rights Reserved
#
# This file is part of iterate.
#
# iterate is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# iterate is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with iterate; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
##################################################################
"""
$Id: test_iterate.py 1595 2006-08-24 00:15:21Z hazmat $
"""

import os, sys

from zope.component import getMultiAdapter

from Acquisition import aq_base
from AccessControl import getSecurityManager

from plone.app.iterate.interfaces import ICheckinCheckoutPolicy

from Products.PloneTestCase import PloneTestCase
from Testing.ZopeTestCase import FunctionalDocFileSuite
PloneTestCase.setupPloneSite()

class TestIterations(PloneTestCase.PloneTestCase):

    def afterSetUp(self):
        self.setRoles(['Manager',])
        
        # Since we depend on ZCML being loaded, we can't do this
        # until the layer is set up
        
        self.portal.portal_quickinstaller.installProduct('plone.app.iterate')

        # add a folder with two documents in it
        self.portal.invokeFactory('Folder', 'docs')
        self.portal.docs.invokeFactory('Document', 'doc1')
        self.portal.docs.invokeFactory('Document', 'doc2')

        # add a working copy folder
        self.portal.invokeFactory('Folder', 'workarea')

        self.repo = self.portal.portal_repository
        self.wf   = self.portal.portal_workflow

    def beforeTearDown(self):
        self.repo = None
        self.wf   = None

    def shim_test( self, test_method):

        try:
            test_method()
        except:
            import sys, pdb, traceback
            ec, e, tb = sys.exc_info()
            traceback.print_exc()            
            pdb.post_mortem( tb )

    def test_workflowState( self ):
        # ensure baseline workflow state is retained on checkin, including security

        doc = self.portal.docs.doc1
        
        # sanity check that owner can edit visible docs
        self.setRoles(['Owner',])
        self.assertTrue( getSecurityManager().checkPermission( "Modify portal content",
                                                               self.portal.docs.doc1 ) )

        self.setRoles(['Manager',])        
        self.wf.doActionFor( doc, 'publish')
        state = self.wf.getInfoFor( doc, 'review_state')
        
        self.repo.save( doc )
        wc = ICheckinCheckoutPolicy( doc ).checkout( self.portal.workarea )
        wc_state = self.wf.getInfoFor( wc, 'review_state')
        
        self.assertNotEqual( state, wc_state )

        ICheckinCheckoutPolicy( wc ).checkin( "modified" )
        bstate = self.wf.getInfoFor( wc, 'review_state')
        self.assertEqual( state, bstate )
        self.setRoles(['Owner',])       

    def test_baselineVersionCreated( self ):
        # if a baseline has no version ensure that one is created on checkout

        doc = self.portal.docs.doc1
        self.assertTrue( self.repo.isVersionable( doc ) )

        history = self.repo.getHistory( doc )
        self.assertEqual( len(history), 0 )

        ICheckinCheckoutPolicy( doc ).checkout( self.portal.workarea )

        history = self.repo.getHistory( doc )
        self.assertEqual( len(history), 1 )

        doc2 = self.portal.docs.doc2
        self.repo.save( doc2 )

        ICheckinCheckoutPolicy( doc2 ).checkout( self.portal.workarea )

        history = self.repo.getHistory( doc2 )
        self.assertEqual( len(history), 1 )
    
    def test_wcNewForwardReferencesCopied( self ):
        # ensure that new wc references are copied back to the baseline on checkin
        doc = self.portal.docs.doc1
        doc.addReference( self.portal.docs )
        self.assertEqual( len(doc.getReferences("zebra")), 0)
        wc = ICheckinCheckoutPolicy( doc ).checkout( self.portal.workarea )
        wc.addReference( self.portal.docs.doc2, "zebra")        
        doc = ICheckinCheckoutPolicy( wc ).checkin( "updated" )
        self.assertEqual( len(doc.getReferences("zebra")), 1 )
        
    def test_wcNewBackwardReferencesCopied( self ):
        # ensure that new wc back references are copied back to the baseline on checkin

        doc = self.portal.docs.doc1
        self.assertEqual( len(doc.getBackReferences("zebra")), 0)
        wc = ICheckinCheckoutPolicy( doc ).checkout( self.portal.workarea )
        self.portal.docs.doc2.addReference( wc, "zebra")
        self.assertEqual( len( wc.getBackReferences("zebra")), 1 )        
        doc = ICheckinCheckoutPolicy( wc ).checkin( "updated")
        self.assertEqual( len( doc.getBackReferences("zebra")), 1 )

    def test_baselineReferencesMaintained( self ):
        # ensure that baseline references are maintained when the object is checked in
        # copies forward, bkw are not copied, but are maintained.

        doc = self.portal.docs.doc1
        doc.addReference( self.portal.docs, "elephant" )
        self.portal.docs.doc2.addReference( doc )

        wc = ICheckinCheckoutPolicy( doc ).checkout( self.portal.workarea )

        doc = ICheckinCheckoutPolicy( wc ).checkin( "updated" )

        self.assertEqual( len(doc.getReferences()), 1 )
        self.assertEqual( len(doc.getBackReferences()), 1 )

    def test_baselineNoCopyReferences( self ):
        # ensure that custom state is maintained with the no copy adapter

        # setup the named ref adapter
        from zope import component
        from Products.Archetypes.interfaces import IBaseObject
        from plone.app.iterate import relation, interfaces
        from plone.app.iterate.tests.utils import CustomReference
        
        component.provideAdapter( 
            adapts = (IBaseObject,),
            provides = interfaces.ICheckinCheckoutReference,
            factory = relation.NoCopyReferenceAdapter,
            name="zebra")

        doc = self.portal.docs.doc1
        ref = doc.addReference( self.portal.docs, "zebra", referenceClass=CustomReference )
        ref.custom_state = "hello world"

        wc = ICheckinCheckoutPolicy( doc ).checkout( self.portal.workarea )

        self.assertEqual( len(wc.getReferences("zebra")), 0)

        doc = ICheckinCheckoutPolicy( wc ).checkin( "updated" )

        self.assertEqual( len(doc.getReferences("zebra")), 1)

        ref = doc.getReferenceImpl("zebra")[0]

        self.assert_( hasattr( ref, "custom_state") )
        self.assertEqual( ref.custom_state, "hello world")


def test_suite():
    from unittest import TestSuite, makeSuite
    suite = TestSuite()
    suite.addTest(makeSuite(TestIterations))
    suite.addTest(FunctionalDocFileSuite(
        'browser.txt',
        test_class=PloneTestCase.FunctionalTestCase))
    return suite



##################################################################
#
# (C) Copyright 2006-2007 ObjectRealms, LLC
# All Rights Reserved
#
# This file is part of iterate.
#
# iterate is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# iterate is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with CMFDeployment; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
##################################################################
"""
A Default Checkin Checkout Policy For Content

$Id: policy.py 1824 2007-02-08 17:59:41Z hazmat $
"""

from zope import component
from zope.event import notify
from zope.interface import implements
from zope.annotation.interfaces import IAnnotations

from Acquisition import Implicit, aq_base, aq_inner, aq_parent

import interfaces
import event
import lock

from relation import WorkingCopyRelation

class CheckinCheckoutPolicyAdapter( object ):
    """
    Default Checkin Checkout Policy For Content
    
    on checkout context is the baseline
    
    on checkin context is the working copy
    """
    
    implements( interfaces.ICheckinCheckoutPolicy )
    component.adapts( interfaces.IIterateAware )

    # used when creating baseline version for first time
    default_base_message = "Created Baseline"
    
    def __init__( self, context ):
        self.context = context

    def checkout( self, container ):
        # see interface
        notify( event.BeforeCheckoutEvent( self.context ) )

        # use the object copier to checkout the content to the container
        copier = component.queryAdapter( self.context, interfaces.IObjectCopier )
        working_copy, relation = copier.copyTo( container )

        # publish the event for any subscribers
        notify( event.CheckoutEvent( self.context, working_copy, relation ) )
       
        # finally return the working copy 
        return working_copy

    def checkin( self, checkin_message ):
        # see interface

        # get the baseline for this working copy, raise if not found
        baseline = self._getBaseline()

        # get a hold of the relation object
        wc_ref = self.context.getReferenceImpl( WorkingCopyRelation.relationship )[ 0]
        
        # publish the event for subscribers, early because contexts are about to be manipulated
        notify( event.CheckinEvent( self.context, baseline, wc_ref, checkin_message ) )

        # merge the object back to the baseline with a copier
        copier = component.queryAdapter( self.context, interfaces.IObjectCopier )
        new_baseline = copier.merge()

        # don't need to unlock the lock disappears with old baseline deletion
        notify( event.AfterCheckinEvent( new_baseline, checkin_message ) )
        
        return new_baseline

    def cancelCheckout( self ):
        # see interface
        
        # get the baseline
        baseline = self._getBaseline()

        # publish an event
        notify( event.CancelCheckoutEvent( self.context, baseline ) )
        
        # delete the working copy        
        wc_container =  aq_parent( aq_inner( self.context ) )
        wc_container.manage_delObjects( [ self.context.getId() ] )

        return baseline

    #################################
    ## Checkin Support Methods

    def _getBaseline( self ):
        # follow the working copy's reference back to the baseline
        refs = self.context.getRefs( WorkingCopyRelation.relationship )

        if not len(refs) == 1:
            raise interfaces.CheckinException( "Baseline count mismatch" )

        if not refs or refs[0] is None:
            raise interfaces.CheckinException( "Baseline has disappeared" )

        baseline = refs[0]
        return baseline


from setuptools import setup, find_packages
import os

version = '1.2.2'

setup(name='plone.app.iterate',
      version=version,
      description="check-out/check-in staging for Plone",
      long_description=\
          open(os.path.join("plone", "app", "iterate", "README.txt")).read() + "\n" + \
          open(os.path.join("docs", "HISTORY.txt")).read(),
      classifiers=[
        "Framework :: Zope2",
        "Framework :: Zope3",
        "Programming Language :: Python",
        "Topic :: Software Development :: Libraries :: Python Modules",
        ],
      keywords='',
      author='Kapil Thangavelu',
      author_email='plone-developers@lists.sourceforge.net',
      url='http://svn.plone.org/svn/plone/plone.app.iterate',
      license='GPL',
      packages=find_packages(exclude=['ez_setup']),
      namespace_packages = ['plone', 'plone.app'],
      include_package_data=True,
      zip_safe=False,
      install_requires=[
          'setuptools',
          "plone.locking",
      ],
      )



##################################################################
#
# (C) Copyright 2006 ObjectRealms, LLC
# All Rights Reserved
#
# This file is part of iterate.
#
# iterate is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# iterate is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with CMFDeployment; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
##################################################################

from Products.CMFEditions import Permissions
from Products.CMFCore.permissions import setDefaultRoles

CheckinPermission  = "iterate : Check in content"
CheckoutPermission = "iterate : Check out content"

setDefaultRoles(CheckinPermission, ('Manager', 'Owner'))
setDefaultRoles(CheckoutPermission, ('Manager', 'Owner'))